/*
 * Copyright 2016 Kalle Olavi Niemitalo.
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#define _GNU_SOURCE 1

#include <mach.h>
#include <hurd.h>
#include <error.h>
#include <stdio.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <mach/mig_support.h>
#include <limits.h>
#include "leaky.h"

static int run_as_child (mach_port_t receive_right);
static int run_as_parent (int child_pid, mach_port_t receive_right);

int
main (void)
{
  kern_return_t kret;
  mach_port_t receive_right = MACH_PORT_NULL;
  mach_port_t send_right = MACH_PORT_NULL;
  mach_msg_type_name_t send_type = MACH_MSG_TYPE_POLYMORPHIC;
  int fd = -1;
  pid_t child_pid = -1;
  
  kret = mach_port_allocate (mach_task_self (), MACH_PORT_RIGHT_RECEIVE,
			     &receive_right);
  if (__hurd_fail (kret))
    error (EXIT_FAILURE, errno,
	   "mach_port_allocate (RECEIVE) error %d", (int) kret);
  printf ("receive_right = %d\n", (int) receive_right);

  child_pid = fork ();
  if (child_pid == -1)
    error (EXIT_FAILURE, errno, "fork");

  if (child_pid == 0)
    return run_as_child (receive_right);
  else
    return run_as_parent (child_pid, receive_right);
}

static int
run_as_child (mach_port_t receive_right)
{
  kern_return_t kret;
  task_t parent_task = MACH_PORT_NULL;
  pid_t parent_pid = -1;
  mach_port_t send_right = MACH_PORT_NULL;
  mach_port_type_t send_type = MACH_MSG_TYPE_POLYMORPHIC;

  printf ("child: pid = %d\n", (int) getpid ());

  /* fork allocated a receive right in the child process, but it's not
     for the same port as in the parent process.  Deallocate that.  */
  kret = mach_port_mod_refs (mach_task_self(), receive_right,
			     MACH_PORT_RIGHT_RECEIVE, -1);
  if (__hurd_fail (kret))
    error (EXIT_FAILURE, errno, "child: mach_port_deallocate error %d", (int) kret);

  parent_pid = getppid ();
  printf ("child: parent_pid = %d\n", (int) parent_pid);

  parent_task = pid2task (parent_pid);
  if (parent_task == MACH_PORT_NULL)
    error (EXIT_FAILURE, errno, "child: pid2task");
  printf ("child: parent_task = %d\n", (int) parent_task);

  kret = mach_port_extract_right (parent_task, receive_right,
				  MACH_MSG_TYPE_MAKE_SEND,
				  &send_right, &send_type);
  if (__hurd_fail (kret))
    error (EXIT_FAILURE, errno, "mach_port_extract_right error %d", (int) kret);
  printf ("child: send_right = %d, send_type = %d\n",
	  (int) send_right, (int) send_type);
  assert (send_type == MACH_MSG_TYPE_PORT_SEND);

  while (1)
    {
      kernel_boot_info_t dummy;
      mach_port_array_t names = NULL;
      mach_msg_type_number_t namesCnt = 0;
      mach_port_type_array_t types = NULL;
      mach_msg_type_number_t typesCnt = 0;
      
      kret = mach_port_names (mach_task_self (),
			      &names, &namesCnt,
			      &types, &typesCnt);
      if (__hurd_fail (kret))
	error (EXIT_FAILURE, errno,
	       "child: mach_port_names error %d", (int) kret);
      printf ("child: namesCnt = %d, typesCnt = %d\n",
	      (int) namesCnt, (int) typesCnt);
      mig_deallocate ((vm_address_t) names, namesCnt * sizeof names[0]);
      mig_deallocate ((vm_address_t) types, typesCnt * sizeof types[0]);

      sleep (1);
      printf ("child: calling\n");
      kret = leaky_probe (send_right, dummy);
      if (__hurd_fail (kret))
	error (errno == EIEIO ? EXIT_FAILURE : 0, errno,
	       "child: leaky_probe error %d", (int) kret);
    }
}

static int
run_as_parent (int child_pid, mach_port_t receive_right)
{
  struct
  {
    mach_msg_header_t header;
    union
    {
      char request[1000];
      struct
      {
	mach_msg_type_t type;
	mach_port_t ports[10];
      } reply;
    } u;
  } msg;
  mach_msg_option_t send_option = MACH_MSG_OPTION_NONE;

  printf ("parent: pid = %d\n", (int) getpid ());

  while (1)
    {
      mach_port_t remote_port;
      mach_msg_type_name_t remote_type;

      mach_msg_return_t mret = mach_msg (
	&msg.header,
	MACH_RCV_MSG | send_option,
	send_option ? msg.header.msgh_size : 0,
	sizeof msg,
	receive_right,
	MACH_MSG_TIMEOUT_NONE,
	MACH_PORT_NULL);
      if (__hurd_fail (mret))
	error (EXIT_FAILURE, errno, "parent: mach_msg error %d\n", (int) mret);

      send_option = MACH_MSG_OPTION_NONE;

      remote_port = msg.header.msgh_remote_port;
      remote_type = MACH_MSGH_BITS_REMOTE (msg.header.msgh_bits);

      printf ("parent: msgh_size = %d, msgh_id = %d, msgh_remote_port = %d\n",
	      (int) msg.header.msgh_size,
	      (int) msg.header.msgh_id,
	      (int) remote_port);
      assert (msg.header.msgh_local_port == receive_right);

      msg.header.msgh_remote_port = MACH_PORT_NULL;
      mach_msg_destroy (&msg.header);

      if (MACH_PORT_VALID (remote_port))
	{
	  int i;

	  memset (&msg.header, 0, sizeof msg.header);
	  msg.header.msgh_bits = MACH_MSGH_BITS (remote_type, 0)
	    | MACH_MSGH_BITS_COMPLEX;
	  msg.header.msgh_size
	    = (char *) (&msg.u.reply + 1) - (char *) &msg;
	  msg.header.msgh_remote_port = remote_port;
	  msg.header.msgh_local_port = MACH_PORT_NULL;
	  msg.header.msgh_id = 0;

	  memset (&msg.u.reply.type, 0, sizeof msg.u.reply.type);
	  msg.u.reply.type.msgt_name = MACH_MSG_TYPE_MOVE_RECEIVE;
	  msg.u.reply.type.msgt_size = sizeof (mach_port_t) * CHAR_BIT;
	  msg.u.reply.type.msgt_number
	    = sizeof msg.u.reply.ports / sizeof msg.u.reply.ports[0];
	  msg.u.reply.type.msgt_inline = 1;
	  msg.u.reply.type.msgt_longform = 0;
	  msg.u.reply.type.msgt_deallocate = 0;
	  msg.u.reply.type.msgt_unused = 0;

	  for (i = 0; i < msg.u.reply.type.msgt_number; i++)
	    {
	      kern_return_t kret = mach_port_allocate (
		mach_task_self (), MACH_PORT_RIGHT_RECEIVE,
		&msg.u.reply.ports[i]);
	      if (__hurd_fail (kret))
		error (EXIT_FAILURE, errno,
		       "parent: mach_port_allocate (RECEIVE) error %d",
		       (int) kret);
	    }

	  send_option = MACH_SEND_MSG;
	}
    }
}
